@@ -17,7 +17,7 @@ from formant.protos.model.v1 import (
 )
 from typing_extensions import Literal
 
-from .exceptions import handle_agent_exceptions
+from .exceptions import InvalidArgument, handle_agent_exceptions
 from .cancellable_stream_thread import CancellableStreamThread
 
 
@@ -42,13 +42,21 @@ def validate_type(
 
 class Client:
     """
-    A client for interacting with the Formant Agent. There are methods for:
-    - Ingesting telemetry datapoints
-    - Ingesting transform frames
-    - Creating events
-    - Handling commands
-    - Reading application configuration
+    A client for interacting with the Formant agent.
     Automatically handles connection and reconnection to the agent.
+    There are methods for:
+
+    * Ingesting telemetry datapoints
+
+    * Creating events
+
+    * Handling commands
+
+    * Ingesting transform frames
+
+    * Reading application configuration
+
+    * Handling teleop control datapoints
     """
 
     def __init__(
@@ -58,7 +66,15 @@ class Client:
         ignore_throttled=False,  # type: bool
         ignore_unavailable=False,  # type: bool
     ):
-
+        """
+        __init__()
+        :param agent_url: The address of the Formant agent API
+        :param enable_logging: If True, this client will log some information to stdout
+        :param ignore_throttled: If True, telemetry datapoint throttle errors
+            will not raise Exceptions. Throttled datapoints are still valid for teleop
+        :param ignore_unavailable: If True, Formant agent unavailable errors
+            will not raise Exceptions.
+        """
         if enable_logging:
             self.logger = logging.getLogger("formant")  # type: Optional[logging.Logger]
             self.logger.setLevel(logging.INFO)
@@ -81,6 +97,9 @@ class Client:
         self._teleop_callback_streams = (
             {}
         )  # type: Dict[Callable[..., None], CancellableStreamThread]
+        self._teleop_heartbeat_callback_streams = (
+            {}
+        )  # type: Dict[Callable[..., None], CancellableStreamThread]
 
         self._setup_agent_communication()
 
@@ -95,7 +114,26 @@ class Client:
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
-        """Post a text datapoint to a telemetry stream."""
+        """
+        Post a text datapoint to a stream.
+
+        :param stream: The name of the Formant stream to post the datapoint on
+        :param value: The text datapoint value
+        :param tags: Tags to include on the posted datapoint
+        :param timestamp: Unix timestamp for the posted datapoint.
+            Uses the current time by default
+
+        .. highlight:: python
+        .. code-block:: python
+
+            from formant.sdk.agent.v1 import Client
+
+            fclient = Client()
+            fclient.post_text(
+                "example.text",
+                "Processed 9 items"
+            )
+        """
 
         validate_type(value, [str])
 
@@ -117,7 +155,15 @@ class Client:
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
-        """Post a JSON datapoint to a telemetry stream."""
+        """
+        Post a JSON datapoint to a telemetry stream.
+
+        :param stream: The name of the Formant stream to post the datapoint on
+        :param value: The encoded JSON datapoint value
+        :param tags: Tags to include on the posted datapoint
+        :param timestamp: Unix timestamp for the posted datapoint.
+            Uses the current time by default
+        """
 
         validate_type(value, [str])
 
@@ -139,7 +185,15 @@ class Client:
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
-        """Post a numeric datapoint to a telemetry stream."""
+        """
+        Post a numeric datapoint to a telemetry stream.
+
+        :param stream: The name of the Formant stream to post the datapoint on
+        :param value: The numeric datapoint value
+        :param tags: Tags to include on the posted datapoint
+        :param timestamp: Unix timestamp for the posted datapoint.
+            Uses the current time by default
+        """
 
         validate_type(value, [float, int])
 
@@ -161,7 +215,32 @@ class Client:
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
-        """Post a numeric set datapoint to a telemetry stream."""
+        """
+        Post a numeric set datapoint to a telemetry stream.
+        Numeric sets are collections of related numeric datapoints.
+
+        :param stream: The name of the Formant stream to post the datapoint on
+        :param numerics_dict: The numeric set datapoint value,
+            a dictionary mapping names to (numeric value, units) tuples.
+        :param tags: Tags to include on the posted datapoint
+        :param timestamp: Unix timestamp for the posted datapoint.
+            Uses the current time by default
+
+        .. highlight:: python
+        .. code-block:: python
+
+            from formant.sdk.agent.v1 import Client
+
+            fclient = Client()
+            fclient.post_numericset(
+                "example.numericset2",
+                {
+                    "frequency": (998, "Hz"),
+                    "usage": (30, "percent"),
+                    "warp factor": (6.0, None),
+                },
+            )
+        """
 
         validate_type(numerics_dict, [dict])
 
@@ -192,21 +271,37 @@ class Client:
     def post_image(
         self,
         stream,  # type: str
-        # raw bytes:
-        value,  # type: bytes
-        content_type="image/jpg",  # type: Literal["image/jpg", "image/png"]
+        value=None,  # type: Optional[bytes]
+        url=None,  # type: Optional[str]
+        content_type="image/jpg",  # type: Literal["image/jpg", "image/png", "video/h264"] # noqa: E501
         tags=None,  # type: Optional[Dict[str, str]]
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
-        """Post an image datapoint to a telemetry stream."""
+        """
+        Post an image datapoint to a telemetry stream.
 
-        validate_type(value, [bytes])
+        :param stream: The name of the Formant stream to post the datapoint on
+        :param value: The datapoint value: raw bytes of an encoded image or frame
+        :param url: The datapoint url: path to local file or valid remote URL for
+             remote files
+        :param content_type: The format of the encoded image or frame.
+            Defaults to "image/jpg".
+        :param tags: Tags to include on the posted datapoint
+        :param timestamp: Unix timestamp for the posted datapoint.
+            Uses the current time by default
+        """
+        if value is not None and url is not None:
+            raise InvalidArgument("Only one of [url, value] can be used.")
+        if value is not None:
+            validate_type(value, [bytes])
+        if url is not None:
+            validate_type(url, [str])
 
         self.agent_stub.PostData(
             datapoint_pb2.Datapoint(
                 stream=stream,
-                image=media_pb2.Image(raw=value, content_type=content_type),
+                image=media_pb2.Image(raw=value, url=url, content_type=content_type),
                 tags=tags if tags else {},
                 timestamp=timestamp if timestamp else current_timestamp(),
             )
@@ -221,7 +316,31 @@ class Client:
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
-        """Post a bitset datapoint to a telemetry stream."""
+        """
+        Post a bitset datapoint to a telemetry stream.
+        A bitset is a collection of related boolean states.
+
+        :param stream: The name of the Formant stream to post the datapoint on
+        :param bitset_dict: The datapoint value, a dictionary mapping names to booleans
+        :param tags: Tags to include on the posted datapoint
+        :param timestamp: Unix timestamp for the posted datapoint.
+            Uses the current time by default
+
+        .. highlight:: python
+        .. code-block:: python
+
+            from formant.sdk.agent.v1 import Client
+
+            fclient = Client()
+            fclient.post_bitset(
+                "example.bitset",
+                {
+                    "standing": False,
+                    "walking": False,
+                    "sitting": True
+                }
+            )
+        """
 
         validate_type(bitset_dict, [dict])
 
@@ -250,7 +369,16 @@ class Client:
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
-        """Post a geolocation datapoint to a telemetry stream."""
+        """
+        Post a geolocation datapoint to a telemetry stream.
+
+        :param stream: The name of the Formant stream to post the datapoint on
+        :param latitude: The datapoint value's latitude
+        :param longitude: The datapoint value's longitude
+        :param tags: Tags to include on the posted datapoint
+        :param timestamp: Unix timestamp for the posted datapoint.
+            Uses the current time by default
+        """
 
         for _ in [latitude, longitude]:
             validate_type(_, [float, int])
@@ -278,7 +406,18 @@ class Client:
         timestamp=None,  # type: Optional[int]
     ):
         # type: (...) -> None
-        """Post a battery datapoint to a telemetry stream."""
+        """
+        Post a battery datapoint to a telemetry stream. Only percentage is required.
+
+        :param stream: The name of the Formant stream to post the datapoint on
+        :param percentage: The battery charge percentage
+        :param voltage: The battery voltage
+        :param current: The battery current
+        :param charge: The battery charge
+        :param tags: Tags to include on the posted datapoint
+        :param timestamp: Unix timestamp for the posted datapoint.
+            Uses the current time by default
+        """
 
         for _ in [percentage, voltage, current, charge]:
             if _:
@@ -318,7 +457,19 @@ class Client:
         rw,  # type: Union[int, float]
     ):
         # type: (...) -> None
-        """Adds a transform frame, used to position datapoints in 3D space."""
+        """
+        Adds a transform frame, used to position datapoints in 3D space.
+
+        :param parent_frame: The parent frame of the posted transform
+        :param child_frame: The child frame of the posted transform
+        :param tx: x-translation
+        :param ty: y-translation
+        :param tz: z-translation
+        :param rx: x-rotation (quaternion)
+        :param ry: y-rotation (quaternion)
+        :param rz: z-rotation (quaternion)
+        :param rw: w-rotation (quaternion)
+        """
 
         for _ in [parent_frame, child_frame]:
             validate_type(_, [str])
@@ -343,13 +494,39 @@ class Client:
     def create_event(
         self,
         message,  # type: str
-        notify=False,  # type: bool
         tags=None,  # type: Optional[Dict[str, str]]
         timestamp=None,  # type: Optional[int]
         end_timestamp=None,  # type: Optional[int]
+        notify=False,  # type: bool
         severity="info",  # type: Literal["info", "warning", "critical", "error"]
     ):
-        """Creates and ingests an event."""
+        # type: (...) -> None
+        """
+        Creates and ingests an event.
+
+        :param message: The text payload of the event
+        :param tags: Tags to include on the event
+        :param timestamp: Unix starting timestamp for the event.
+            Uses the current time by default
+        :param end_timestamp: Unix ending timestamp for the event.
+            Must be greater than timestamp.
+            If end_timestamp is supplied, the event will span a length of time
+        :param notify: If True, the created event will trigger a Formant notification
+        :param severity: The severity level of the event
+
+        .. highlight:: python
+        .. code-block:: python
+
+            from formant.sdk.agent.v1 import Client
+
+            fclient = Client()
+            fclient.create_event(
+                "Confinement beam to warp frequency 0.4e17 hz",
+                tags={"Region": "North"},
+                notify=True,
+                severity="warning"
+            )
+        """
 
         validate_type(message, [str])
 
@@ -381,10 +558,14 @@ class Client:
     ):
         # type: (...) -> Optional[commands_pb2.CommandRequest]
         """
-        If there is a command request in the agent's queue whose "command" value matches
-        an element of the given command filter, takes and returns the command request.
+        If there is a command request in the agent's queue whose "command" value
+        matches an element of the given command filter, takes and returns the
+        command request.
         Otherwise, returns None if there are no matching command requests
         in the agent's queue.
+
+        :param command_filter: A list of command names.
+            This method only returns commands whose names are in this list.
         """
         command_request_request = agent_pb2.GetCommandRequestRequest(
             command_filter=command_filter
@@ -407,6 +588,10 @@ class Client:
         """
         Sends a command response for an identified command request to Formant.
         Returns an error if there was a problem sending the command response.
+
+        :param request_id: The ID of the command request to which this method responds
+        :param request_id: Whether the command was successfully executed
+        :param datapoint: A datapoint related to the command
         """
         response = commands_pb2.CommandResponse(
             request_id=request_id, success=success, datapoint=datapoint,
@@ -419,62 +604,128 @@ class Client:
         f,  # type: Callable[[commands_pb2.CommandRequest], None]
         command_filter=None,  # type: Optional[List[str]]
     ):
-        # type : (...) -> None
+        # type: (...) -> None
         """
         Command requests issued to the agent whose "command" value matches an element
         of the given command filter will be streamed into the provided callback.
         If no command filter is provided, all command requests will be handled.
+
+        :param f: A callback that will be executed on command requests
+            as they are received by the Formant agent.
+        :param command_filter: A list of command names. The provided callback
+            is only exectued on commands whose names are in this list
         """
 
         def create_stream():
             return self._get_command_request_stream(command_filter)
 
         self._command_request_callback_streams[f] = CancellableStreamThread(
-            f, create_stream, self.logger, attribute="request"
+            f,
+            create_stream,
+            self.logger,
+            attribute="request",
+            ignore_unavailable=self.ignore_unavailable,
         )
 
     def unregister_command_request_callback(
         self, f,  # type: Callable[[commands_pb2.CommandRequest], None]
     ):
-        # type : (...) -> None
+        # type: (...) -> None
         """
         Unregisters previously registered command request callback.
+
+        :param f: The command request callback to be unregistered
         """
         stream = self._command_request_callback_streams.pop(f, None)
         if stream is not None:
             stream.cancel()
 
-    # Receive teleop control data
+    # Teleop
 
     def register_teleop_callback(
         self,
-        f,  # type: Callable[[datapoint_pb2.Datapoint], None]
+        f,  # type: Callable[[datapoint_pb2.ControlDatapoint], None]
         stream_filter=None,  # type: Optional[List[str]]
     ):
-        # type : (...) -> None
+        # type: (...) -> None
         """
         Control datapoints received from teleop whose "stream" value matches an element
         of the given stream filter will be streamed into the provided callback.
-        If no stream filter is provided, all control datapoints will be received.
+        If no stream filter is provided, control datapoints from all streams
+        will be received.
+
+        :param f: A callback that will be executed on teleop control datapoints as they
+            are received by the Formant agent
+        :param stream_filter: A list of stream names. The provided callback
+            is only exectued on control datapoints whose names are in this list
         """
 
         def create_stream():
             return self._get_teleop_stream(stream_filter)
 
         self._teleop_callback_streams[f] = CancellableStreamThread(
-            f, create_stream, self.logger, attribute="control_datapoint"
+            f,
+            create_stream,
+            self.logger,
+            attribute="control_datapoint",
+            ignore_unavailable=self.ignore_unavailable,
         )
 
     def unregister_teleop_callback(
-        self, f  # type: Callable[[datapoint_pb2.Datapoint], None]
+        self, f  # type: Callable[[datapoint_pb2.ControlDatapoint], None]
     ):
+        # type: (...) -> None
         """
         Unregisters previously registered teleop callback.
+
+        :param f: The teleop callback to be unregistered
         """
         stream = self._teleop_callback_streams.pop(f, None)
         if stream is not None:
             stream.cancel()
 
+    @handle_agent_exceptions
+    def get_teleop_info(self):
+        # type: (...) -> agent_pb2.GetTeleopInfoResponse
+        """
+        Returns current information about teleop connection count.
+        """
+        request = agent_pb2.GetTeleopInfoRequest()
+        return self.agent_stub.GetTeleopInfo(request)
+
+    def register_teleop_heartbeat_callback(
+        self, f,  # type: Callable[[agent_pb2.GetTeleopHeartbeatStreamResponse], None]
+    ):
+        # type : (...) -> None
+        """
+        The provided callback will be called once each time a heartbeat
+        is received over Formant teleop. Heartbeats are streamed from the
+        operator machine at 20Hz on a UDP-like channel. This method can be used
+        to quickly detect teleop disconnections.
+
+        :param f: A callback that will be called when a heartbeat is received
+        """
+
+        def create_stream():
+            return self._get_teleop_heartbeat_stream()
+
+        self._teleop_heartbeat_callback_streams[f] = CancellableStreamThread(
+            f, create_stream, self.logger, ignore_unavailable=self.ignore_unavailable
+        )
+
+    def unregister_teleop_heartbeat_callback(
+        self, f  # type: Callable[[], None]
+    ):
+        # type : (...) -> None
+        """
+        Unregisters previously registered teleop heartbeat callback.
+
+        :param f: The teleop heartbeat callback to be unregistered
+        """
+        stream = self._teleop_heartbeat_callback_streams.pop(f, None)
+        if stream is not None:
+            stream.cancel()
+
     # Reading application configuration
 
     def get_app_config(self, key, *args):
@@ -483,14 +734,17 @@ class Client:
         Returns the value for the given key that was set in
         Formant application configuration for this device,
         or returns the given default value.
+
+        :param key: The application configuration key
+        :param args: (One additional argument)
+            The default value to return if the key is not found.
         """
         if len(args) > 1:
             raise TypeError("function takes at most two args: (key: str, default: Any)")
         default = args[0] if len(args) == 1 else None
 
-        if len(self._app_config.keys()) == 0:
-            self._update_application_configuration()
-
+        # update if there's a connection to the agent API
+        self._update_application_configuration()
         return self._app_config.get(key, default)
 
     @handle_agent_exceptions
@@ -507,6 +761,8 @@ class Client:
         """
         Adds a function to the list of callbacks that are executed by the client
         when this device receives updated configuration from Formant.
+
+        :param f: The configuration update callback to be registered
         """
         self._config_update_callbacks.append(f)
 
@@ -515,18 +771,11 @@ class Client:
         """
         Removes a function from the list of callbacks that are executed by the client
         when this device receives updated configuration from Formant.
+
+        :param f: The configuration update callback to be unregistered
         """
         self._config_update_callbacks.remove(f)
 
-    @handle_agent_exceptions
-    def get_teleop_info(self):
-        # type: (...) -> agent_pb2.GetTeleopInfoResponse
-        """
-        Returns current information about teleop connection count.
-        """
-        request = agent_pb2.GetTeleopInfoRequest()
-        return self.agent_stub.GetTeleopInfo(request)
-
     # Internal methods
 
     def _setup_agent_communication(self):
@@ -565,3 +814,8 @@ class Client:
         return self.agent_stub.GetTeleopControlDataStream(
             agent_pb2.GetTeleopControlDataStreamRequest(stream_filter=stream_filter)
         )
+
+    def _get_teleop_heartbeat_stream(self):
+        return self.agent_stub.GetTeleopHeartbeatStream(
+            agent_pb2.GetTeleopHeartbeatStreamRequest()
+        )
